import { DeepClient } from './client.js';
import { Id, Link } from './minilinks.js';
export interface PackageIdentifier {
    name?: string;
    version?: string;
    uri?: string;
    type?: string;
    options?: any;
}
export interface Package {
    package?: PackageIdentifier;
    data?: Array<PackageItem>;
    dependencies?: Array<PackageIdentifier>;
    strict?: boolean;
    errors?: PackagerError[];
}
export interface PackageItem {
    id: Id | string;
    type?: Id | string;
    from?: Id | string;
    to?: Id | string;
    value?: PackagerValue;
    alias?: Id;
    package?: {
        dependencyId: Id;
        containValue: string;
    };
}
export interface PackagerValue {
    id?: Id;
    link_id?: Id;
    value?: number | string | any;
}
export type PackagerError = any;
export interface PackagerImportResult {
    errors?: PackagerError[];
    ids?: Id[];
    packageId?: Id;
    namespaceId?: Id;
    inserting?: any[];
    updating?: any[];
    pckg?: Package;
}
export type PackagerMutated = {
    [index: number]: boolean;
};
export interface PackagerExportOptions {
    packageLinkId: Id;
}
export interface PackagerLink extends Link<any> {
    value?: any;
}
export declare function sort(pckg: Package, data: any[], errors: PackagerError[], references: {
    id: string;
    from: string;
    to: string;
    type: string;
}): {
    sorted: any[];
};
export declare class Packager<L extends Link<any>> {
    pckg: Package;
    client: DeepClient<any>;
    constructor(client: DeepClient<L>);
    validate(pckg: Package, errors: any[]): void;
    update(packageId: Id, pckg: Package): Promise<PackagerImportResult>;
    import(pckg: Package): Promise<PackagerImportResult>;
    apply(pckg: Package, _packageId?: Id): Promise<{
        inserting?: any[];
        updating?: any[];
        package?: PackageIdentifier;
        ids?: number[];
        errors?: any[];
    }>;
    exec({ pckg, packageId, inserting, updating }: {
        pckg: Package;
        packageId?: Id;
        inserting?: any[];
        updating?: any[];
    }): Promise<{
        packageId: Id;
        namespaceId: any;
    }>;
    export(options: PackagerExportOptions | Id): Promise<Package>;
}
