var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import Debug from 'debug';
import { delay } from './promise.js';
import isEqual from 'lodash/isEqual.js';
const debug = Debug('deeplinks:packager');
const log = debug.extend('log');
const error = debug.extend('error');
const corePackage = '@deep-foundation/core';
export function sort(pckg, data, errors = [], references) {
    let sorted = [];
    if (pckg.strict) {
        sorted = data;
    }
    else {
        for (let i = 0; i < data.length; i++) {
            const item = data[i];
            if (item.value && !item[references.type]) {
                const max = sorted.reduce((p, l, index) => l[references.id] == item[references.id] ? index : p, 0);
                sorted.splice((max) + 1, 0, item);
            }
            else if (item.package) {
                sorted.splice(0, 0, item);
            }
            else {
                const _ = item === null || item === void 0 ? void 0 : item._;
                const firstDependIndex = sorted.findIndex((l, index) => {
                    return l[references.from] == item[references.id] || l[references.to] == item[references.id] || ((l[references.type] == item[references.id]) && !l._);
                });
                const maxDependIndex = sorted.reduce((p, l, index) => {
                    return ((l[references.id] == item[references.type]) && !_) || l[references.id] == item[references.from] || l[references.id] === item[references.to] ? index : p;
                }, 0);
                const _index = (!!~firstDependIndex && firstDependIndex < maxDependIndex ? firstDependIndex : maxDependIndex);
                sorted.splice(_index + 1, 0, item);
            }
        }
    }
    return { sorted };
}
export class Packager {
    constructor(client) {
        this.client = client;
    }
    validate(pckg, errors) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        if (pckg.strict)
            return;
        if (!((_a = pckg === null || pckg === void 0 ? void 0 : pckg.package) === null || _a === void 0 ? void 0 : _a.name))
            errors.push(`!pckg?.package?.name`);
        if (!((_b = pckg === null || pckg === void 0 ? void 0 : pckg.package) === null || _b === void 0 ? void 0 : _b.version))
            errors.push(`!pckg?.package?.version`);
        if (pckg.hasOwnProperty('dependencies')) {
            if (typeof (pckg === null || pckg === void 0 ? void 0 : pckg.dependencies) === 'object') {
                const keys = Object.keys(pckg === null || pckg === void 0 ? void 0 : pckg.dependencies);
                for (let i = 0; i < keys.length; i++) {
                    const dep = pckg === null || pckg === void 0 ? void 0 : pckg.dependencies[keys[i]];
                    if (typeof (dep) === 'object') {
                        if (typeof (dep.name) !== 'string')
                            errors.push(`!dep[${i}].name`);
                    }
                    else
                        errors.push(`!pckg.dependencies[keys[${i}]]`);
                }
            }
            else
                errors.push('!pckg?.dependencies');
        }
        if (Object.prototype.toString.call(pckg === null || pckg === void 0 ? void 0 : pckg.data) === '[object Array]') {
            for (let i = 0; i < (pckg === null || pckg === void 0 ? void 0 : pckg.data.length); i++) {
                const item = pckg === null || pckg === void 0 ? void 0 : pckg.data[i];
                if (typeof (item) === 'object') {
                    if (typeof (item === null || item === void 0 ? void 0 : item.id) !== 'string' && typeof (item === null || item === void 0 ? void 0 : item.id) !== 'number')
                        errors.push(`!item[${i}].id`);
                    if (typeof (item === null || item === void 0 ? void 0 : item.package) === 'object') {
                        if (typeof ((_c = item === null || item === void 0 ? void 0 : item.package) === null || _c === void 0 ? void 0 : _c.dependencyId) !== 'number')
                            errors.push(`!item[${i}].package?.dependencyId`);
                        if (typeof ((_d = item === null || item === void 0 ? void 0 : item.package) === null || _d === void 0 ? void 0 : _d.containValue) !== 'string')
                            errors.push(`!item[${i}].package?.containValue`);
                        if (!((_e = pckg.dependencies) === null || _e === void 0 ? void 0 : _e[item.package.dependencyId]))
                            errors.push(`!pckg.dependencies?.[pckg.data[${i}].package.dependencyId,(${item.package.dependencyId})]`);
                    }
                    else {
                        if (typeof (item === null || item === void 0 ? void 0 : item.type) !== 'string' && typeof (item === null || item === void 0 ? void 0 : item.type) !== 'number')
                            errors.push(`!item[${i}(id:${item === null || item === void 0 ? void 0 : item.id})].type`);
                        if (item.hasOwnProperty('from') && typeof (item === null || item === void 0 ? void 0 : item.from) !== 'undefined')
                            if (typeof (item === null || item === void 0 ? void 0 : item.from) !== 'string' && typeof (item === null || item === void 0 ? void 0 : item.from) !== 'number')
                                errors.push(`!item[${i}(id:${item === null || item === void 0 ? void 0 : item.id})].from`);
                        if (item.hasOwnProperty('to') && typeof (item === null || item === void 0 ? void 0 : item.to) !== 'undefined')
                            if (typeof (item === null || item === void 0 ? void 0 : item.to) !== 'string' && typeof (item === null || item === void 0 ? void 0 : item.to) !== 'number')
                                errors.push(`!item[${i}(id:${item === null || item === void 0 ? void 0 : item.id})].to`);
                        if (item.value) {
                            if (typeof (item === null || item === void 0 ? void 0 : item.value) !== 'object')
                                errors.push(`!item[${i}].value`);
                            else {
                                if (typeof ((_f = item === null || item === void 0 ? void 0 : item.value) === null || _f === void 0 ? void 0 : _f.value) !== 'number' && typeof ((_g = item === null || item === void 0 ? void 0 : item.value) === null || _g === void 0 ? void 0 : _g.value) !== 'string' && typeof ((_h = item === null || item === void 0 ? void 0 : item.value) === null || _h === void 0 ? void 0 : _h.value) !== 'object')
                                    errors.push(`!item[${i}].value.value`);
                            }
                        }
                    }
                }
                else
                    errors.push(`!pckg.dependencies[keys[${i}]]`);
            }
        }
        else
            errors.push('!pckg?.dependencies');
    }
    update(packageId, pckg) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.client.isId(packageId))
                return { errors: ['!packageId'] };
            if (typeof (pckg) !== 'object')
                return { errors: ['!pckg'] };
            const { ids, inserting, updating, errors } = yield this.apply(pckg, packageId);
            if (errors.length)
                return { errors };
            const { namespaceId } = yield this.exec({ inserting, updating, pckg, packageId });
            return { ids, inserting, updating, errors, pckg, packageId, namespaceId };
        });
    }
    import(pckg) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            if (typeof (pckg) !== 'object')
                return { errors: ['!pckg'] };
            const { ids, inserting, updating, errors } = yield this.apply(pckg);
            if (errors.length)
                return { errors };
            const dc = '@deep-foundation/core';
            const Package = yield this.client.id(dc, 'Package');
            const { data: [{ id: packageId }] } = yield this.client.insert({ type_id: Package, string: (_a = pckg === null || pckg === void 0 ? void 0 : pckg.package) === null || _a === void 0 ? void 0 : _a.name });
            const { namespaceId } = yield this.exec({ inserting, updating, packageId, pckg });
            return { ids, inserting, updating, errors, pckg, packageId, namespaceId };
        });
    }
    apply(pckg, _packageId) {
        return __awaiter(this, void 0, void 0, function* () {
            const deep = this.client;
            const errors = [];
            let idI = 0;
            const ids = Array.from(Array(pckg.data.length).keys()).map(id => id + 99999999);
            const dc = '@deep-foundation/core';
            const Contain = yield deep.id(dc, 'Contain');
            const packageId = _packageId || ids[idI++];
            const inserting = [];
            const updating = [];
            const remembered = {};
            const remember = (itemId, linkId) => {
                remembered[itemId] = linkId;
            };
            const remind = (itemId) => __awaiter(this, void 0, void 0, function* () {
                console.log('remind', itemId);
                if (!itemId)
                    return 0;
                if (remembered[itemId])
                    return remembered[itemId];
                else {
                    const item = pckg.data.find(i => i.id == itemId);
                    yield fill(item);
                }
            });
            const fill = (item) => __awaiter(this, void 0, void 0, function* () {
                var _a, _b, _c;
                console.log('fill', item);
                yield delay(100);
                if (item.package) {
                    const pId = yield deep.id(pckg.dependencies[item.package.dependencyId].name, item.package.containValue);
                    console.log('fill', 'package', pId, pckg.dependencies[item.package.dependencyId].name, item.package.containValue);
                    remember(item.id, pId);
                }
                else if (item.alias) {
                    const linkId = yield remind(item.alias);
                    const ins = { type_id: Contain, from_id: packageId, to_id: linkId, string: item.id };
                    inserting.push(ins);
                    console.log('fill', 'inserting alias', ins);
                    remember(item.id, linkId);
                }
                else {
                    let exists;
                    if (_packageId)
                        exists = yield deep.one({ id: { _id: [packageId, item.id] } });
                    if (!exists) {
                        const linkId = ids[idI++];
                        const value = (_a = item === null || item === void 0 ? void 0 : item.value) === null || _a === void 0 ? void 0 : _a.value;
                        const v = typeof (value);
                        const ins = Object.assign(Object.assign(Object.assign(Object.assign({ id: linkId, name: item.id, type_id: yield remind(item.type) }, (item.from ? { from_id: yield remind(item.from) } : {})), (item.to ? { to_id: yield remind(item.to) } : {})), (item.to ? { to_id: yield remind(item.to) } : {})), (!!value ? { [v]: { value } } : {}));
                        inserting.push(ins);
                        console.log('fill', 'inserting !exists', ins);
                        remember(item.id, linkId);
                    }
                    else {
                        const linkId = exists.id;
                        console.log('fill', 'inserting exists', exists);
                        remember(item.id, linkId);
                        const value = (_b = item === null || item === void 0 ? void 0 : item.value) === null || _b === void 0 ? void 0 : _b.value;
                        if (!isEqual((_c = exists === null || exists === void 0 ? void 0 : exists.value) === null || _c === void 0 ? void 0 : _c.value, value))
                            updating.push({ id: linkId, value });
                        const from_id = yield remind(item.from);
                        const to_id = yield remind(item.to);
                        if (from_id != exists.from_id || to_id != exists.to_id)
                            updating.push({ id: linkId, from_id, to_id });
                    }
                }
            });
            try {
                this.validate(pckg, errors);
                if (errors.length)
                    return { errors };
                for (let item of pckg.data) {
                    yield fill(item);
                }
                if (errors.length)
                    return { errors };
                return { ids, errors, inserting, updating, package: pckg.package };
            }
            catch (e) {
                errors.push(deep.stringify(e));
            }
            return { ids, errors };
        });
    }
    exec({ pckg, packageId, inserting, updating }) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            const deep = this.client;
            if (!deep.isId(packageId))
                throw new Error('!packageId');
            yield deep.insert(inserting.map(i => (Object.assign(Object.assign({}, i), { containerId: packageId }))));
            for (let u of updating) {
                if (u.value)
                    yield deep.value(u.id, u.value);
                else
                    yield deep.update(u.id, { from_id: u.from_id, to_id: u.to_id });
            }
            let version;
            const dc = '@deep-foundation/core';
            const PackageVersion = yield deep.id(dc, 'PackageVersion');
            version = yield deep.one({ type_id: PackageVersion, to_id: packageId });
            if (version)
                yield deep.value(version.id, pckg.package.version);
            else {
                const PackageNamespace = yield deep.id(dc, 'PackageNamespace');
                let namespace = yield deep.one({ type_id: PackageNamespace, string: pckg.package.name });
                if (!namespace)
                    namespace = (_b = (_a = (yield deep.insert({ type_id: PackageNamespace, string: pckg.package.name }))) === null || _a === void 0 ? void 0 : _a.data) === null || _b === void 0 ? void 0 : _b[0];
                yield deep.insert({ type_id: PackageVersion, from_id: namespace.id, to_id: packageId, string: pckg.package.version });
                return { packageId, namespaceId: namespace.id };
            }
            return { packageId, namespaceId: version.from_id };
        });
    }
    export(options) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l;
        return __awaiter(this, void 0, void 0, function* () {
            const deep = this.client;
            const packageId = typeof (options) === 'object' ? options === null || options === void 0 ? void 0 : options.packageLinkId : options;
            const errors = [];
            const pckg = {
                package: {},
                errors,
                data: [],
                dependencies: [],
            };
            const error = (message) => {
                errors.push(message);
                throw new Error('!exported');
            };
            try {
                if (!(typeof (packageId) === 'number' || typeof (packageId) === 'string'))
                    errors.push('!packageId');
                if (errors.length)
                    return pckg;
                const dc = '@deep-foundation/core';
                const Contain = yield deep.id(dc, 'Contain');
                const Package = yield deep.id(dc, 'Package');
                const PackageVersion = yield deep.id(dc, 'PackageVersion');
                const { data } = yield deep.select({
                    type_id: Package, id: packageId,
                    return: {
                        _version: {
                            relation: 'in', type_id: PackageVersion,
                        },
                        _contains: {
                            relation: 'out', type_id: Contain,
                            return: {
                                _item: {
                                    relation: 'to'
                                },
                            },
                        },
                    },
                });
                const load = (id) => __awaiter(this, void 0, void 0, function* () {
                    return yield deep.one({
                        type_id: Contain,
                        to_id: id,
                        return: {
                            _package: {
                                relation: 'from',
                                return: {
                                    _version: {
                                        relation: 'in', type_id: PackageVersion,
                                    },
                                }
                            },
                        }
                    });
                });
                if (!(data === null || data === void 0 ? void 0 : data[0]))
                    error(`!package ${packageId}`);
                pckg.package.name = (_b = (_a = data === null || data === void 0 ? void 0 : data[0]) === null || _a === void 0 ? void 0 : _a.value) === null || _b === void 0 ? void 0 : _b.value;
                pckg.package.version = ((_f = (_e = (_d = (_c = data === null || data === void 0 ? void 0 : data[0]) === null || _c === void 0 ? void 0 : _c._version) === null || _d === void 0 ? void 0 : _d[0]) === null || _e === void 0 ? void 0 : _e.value) === null || _f === void 0 ? void 0 : _f.value) || '0.0.0';
                const contains = {};
                const ids = {};
                const packed = {};
                let dependId = 1;
                const depended = {};
                const dependencies = {};
                for (let l of data) {
                    contains[(_g = l === null || l === void 0 ? void 0 : l.value) === null || _g === void 0 ? void 0 : _g.value] = l;
                    ids[l === null || l === void 0 ? void 0 : l.to_id] = l;
                }
                const find = (id) => __awaiter(this, void 0, void 0, function* () {
                    var _m, _o, _p, _q, _r, _s, _t, _u, _v, _w, _x, _y, _z, _0, _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11, _12, _13, _14, _15, _16, _17, _18, _19, _20, _21, _22, _23, _24, _25, _26, _27, _28, _29, _30, _31, _32, _33;
                    if (!ids[id]) {
                        ids[id] = yield load(id);
                        console.log('find', 'id', id, 'loaded', (_o = (_m = ids[id]) === null || _m === void 0 ? void 0 : _m.value) === null || _o === void 0 ? void 0 : _o.value, (_r = (_q = (_p = ids[id]) === null || _p === void 0 ? void 0 : _p._package) === null || _q === void 0 ? void 0 : _q.value) === null || _r === void 0 ? void 0 : _r.value);
                        if (!((_t = (_s = ids[id]) === null || _s === void 0 ? void 0 : _s.value) === null || _t === void 0 ? void 0 : _t.value))
                            errors.push(`!${id}.name`);
                        if (!((_v = (_u = ids[id]) === null || _u === void 0 ? void 0 : _u._package) === null || _v === void 0 ? void 0 : _v.id))
                            error(`!${id}.package`);
                        if (!dependencies.hasOwnProperty(`${(_x = (_w = ids[id]) === null || _w === void 0 ? void 0 : _w._package) === null || _x === void 0 ? void 0 : _x.id}`)) {
                            dependencies[(_z = (_y = ids[id]) === null || _y === void 0 ? void 0 : _y._package) === null || _z === void 0 ? void 0 : _z.id] = pckg.dependencies.length;
                            if (!((_2 = (_1 = (_0 = ids[id]) === null || _0 === void 0 ? void 0 : _0._package) === null || _1 === void 0 ? void 0 : _1.value) === null || _2 === void 0 ? void 0 : _2.value))
                                errors.push(`package !${(_4 = (_3 = ids[id]) === null || _3 === void 0 ? void 0 : _3._package) === null || _4 === void 0 ? void 0 : _4.id}?.value?.value`);
                            if (!((_9 = (_8 = (_7 = (_6 = (_5 = ids[id]) === null || _5 === void 0 ? void 0 : _5._package) === null || _6 === void 0 ? void 0 : _6._version) === null || _7 === void 0 ? void 0 : _7[0]) === null || _8 === void 0 ? void 0 : _8.value) === null || _9 === void 0 ? void 0 : _9.value))
                                errors.push(`package !${(_11 = (_10 = ids[id]) === null || _10 === void 0 ? void 0 : _10._package) === null || _11 === void 0 ? void 0 : _11.id}?.version?.value?.value`);
                            pckg.dependencies.push({
                                name: (_14 = (_13 = (_12 = ids[id]) === null || _12 === void 0 ? void 0 : _12._package) === null || _13 === void 0 ? void 0 : _13.value) === null || _14 === void 0 ? void 0 : _14.value,
                                version: (_19 = (_18 = (_17 = (_16 = (_15 = ids[id]) === null || _15 === void 0 ? void 0 : _15._package) === null || _16 === void 0 ? void 0 : _16._version) === null || _17 === void 0 ? void 0 : _17[0]) === null || _18 === void 0 ? void 0 : _18.value) === null || _19 === void 0 ? void 0 : _19.value
                            });
                            console.log('find', 'id', id, 'add dependency', dependencies[(_21 = (_20 = ids[id]) === null || _20 === void 0 ? void 0 : _20._package) === null || _21 === void 0 ? void 0 : _21.id], '=', (_24 = (_23 = (_22 = ids[id]) === null || _22 === void 0 ? void 0 : _22._package) === null || _23 === void 0 ? void 0 : _23.value) === null || _24 === void 0 ? void 0 : _24.value);
                        }
                        if (!depended.hasOwnProperty(id)) {
                            depended[id] = { id: dependId++, package: {
                                    dependencyId: dependencies[(_26 = (_25 = ids[id]) === null || _25 === void 0 ? void 0 : _25._package) === null || _26 === void 0 ? void 0 : _26.id],
                                    containValue: (_28 = (_27 = ids[id]) === null || _27 === void 0 ? void 0 : _27.value) === null || _28 === void 0 ? void 0 : _28.value,
                                } };
                            pckg.data.push(depended[id]);
                            console.log('find', 'id', id, 'add depended', dependId);
                        }
                    }
                    else {
                        console.log('find', 'id', id, 'exists', (_30 = (_29 = ids[id]) === null || _29 === void 0 ? void 0 : _29.value) === null || _30 === void 0 ? void 0 : _30.value, (_33 = (_32 = (_31 = ids[id]) === null || _31 === void 0 ? void 0 : _31._package) === null || _32 === void 0 ? void 0 : _32.value) === null || _33 === void 0 ? void 0 : _33.value);
                    }
                    if (!ids[id]) {
                        error(`!${id}`);
                    }
                    return ids[id];
                });
                const pack = (contain) => __awaiter(this, void 0, void 0, function* () {
                    var _34, _35, _36, _37;
                    const link = contain._item;
                    console.log('pack', 'contain', contain.id, 'link', link.id);
                    if (packed[link.id])
                        return packed[link.id];
                    let _type, _from, _to;
                    _type = yield find(link.type_id);
                    if (link.from_id)
                        _from = yield find(link.from_id);
                    if (link.to_id)
                        _to = yield find(link.to_id);
                    packed[link.id] = {
                        id: (_34 = contain === null || contain === void 0 ? void 0 : contain.value) === null || _34 === void 0 ? void 0 : _34.value,
                        type: (_type === null || _type === void 0 ? void 0 : _type._package) ? depended[link.type_id].id : (_35 = _type === null || _type === void 0 ? void 0 : _type.value) === null || _35 === void 0 ? void 0 : _35.value,
                        from: (_from === null || _from === void 0 ? void 0 : _from._package) ? depended[link.from_id].id : (_36 = _from === null || _from === void 0 ? void 0 : _from.value) === null || _36 === void 0 ? void 0 : _36.value,
                        to: (_to === null || _to === void 0 ? void 0 : _to._package) ? depended[link.to_id].id : (_37 = _to === null || _to === void 0 ? void 0 : _to.value) === null || _37 === void 0 ? void 0 : _37.value,
                        value: link === null || link === void 0 ? void 0 : link.value,
                    };
                    pckg.data.push(packed[link.id]);
                });
                for (let l of (_h = data === null || data === void 0 ? void 0 : data[0]) === null || _h === void 0 ? void 0 : _h._contains) {
                    if (!((_j = l === null || l === void 0 ? void 0 : l.value) === null || _j === void 0 ? void 0 : _j.value))
                        errors.push(`!${l.id}.name`);
                    contains[(_k = l === null || l === void 0 ? void 0 : l.value) === null || _k === void 0 ? void 0 : _k.value] = l;
                    ids[l === null || l === void 0 ? void 0 : l.to_id] = l;
                }
                for (let l of (_l = data === null || data === void 0 ? void 0 : data[0]) === null || _l === void 0 ? void 0 : _l._contains) {
                    yield pack(l);
                }
                return pckg;
            }
            catch (e) {
                errors.push(deep.stringify(e));
                return { errors };
            }
        });
    }
}
//# sourceMappingURL=packager.js.map